//
// plot_component.cc
// 
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "plot_component.h"
#include <ntk/ntk_with_macros.h>

void PlotComponent :: 
setPlot(const vcl_map<double,double>& points, 
        double xmin, double xmax, double ymin, double ymax)
{
  ntk_assert(xmin <= xmax, "xmin must be smaller than xmax");
  ntk_assert(ymin <= ymax, "ymin must be smaller than ymax");
  m_x_range.first = xmin; m_x_range.second = xmax;
  m_y_range.first = ymin; m_y_range.second = ymax;
  m_plot = points;
  
  // Create Path.
  m_path.clear();
    
  // Points
  if (m_plot.size() > 0)
  {
    vcl_map<double,double>::const_iterator it = m_plot.begin();
    m_path.startNewSubPath (it->first, it->second);
    for (; it != m_plot.end(); ++it)
    {
      m_path.lineTo(it->first, it->second);
    }
  }
  
  repaint();
}

void PlotComponent :: 
paint(Graphics& g)
{
  g.fillAll(Colours::white);
  if (m_plot.size() == 0) return;
  g.saveState();
  g.setColour(Colours::black);
  
  const int left_border = 2;
  const int bottom_border = 2;
  const int text_height = 30;
  const int top_border = 2;
  const double max_plot_y = top_border+text_height/2.0;
  const double xratio = (getWidth()-left_border*2)/(m_x_range.second-m_x_range.first);
  const double yratio = (getHeight()-max_plot_y-bottom_border)/(m_y_range.second-m_y_range.first);
  AffineTransform transform = AffineTransform::scale(xratio, -yratio);
  transform = transform.followedBy(AffineTransform::translation(left_border,getHeight()-bottom_border));
  g.strokePath (m_path, PathStrokeType (1.0f), transform);
  
  // X & Y axis
  double startx=0, starty=0, endx=m_x_range.second-m_x_range.first, endy=0;
  transform.transformPoint(startx, starty);
  transform.transformPoint(endx, endy);
  g.drawLine(startx, starty, endx, endy);
  
  startx=0; starty=0; endx=0; endy=m_y_range.second-m_y_range.first;
  transform.transformPoint(startx, starty);
  transform.transformPoint(endx, endy);
  g.drawLine(startx, starty, endx, endy);
   
  // Y values on axis.
  startx=0; starty=m_y_range.second;
  transform.transformPoint(startx, starty);
  g.drawLine(startx-2, starty, endx+3, endy);
  
  g.drawFittedText(String(m_y_range.second,2), int(startx+8), int(starty-(text_height/2)), 
                   50, text_height, Justification::left, 1);
  
  g.restoreState();
}

void PlotComponent :: 
mouseDown(const MouseEvent& e)
{
  if (!e.mods.isRightButtonDown() || m_plot.size()==0)
    return;

  PopupMenu m;
  m.addItem(1, "Save As Gnuplot data file ...");
  const int result = m.show();

  if (result == 0)
  {
    // user dismissed the menu without picking anything
  }
  else if (result == 1)
  {
    FileChooser chooser("Please select a plot file to save into...",
                        File::getCurrentWorkingDirectory().getChildFile("data.plot"),
                        "*.plot");

    if (chooser.browseForFileToSave(true))
    {
      File f = chooser.getResult();
      if (!f.replaceWithText(""))
        AlertWindow::showMessageBox(AlertWindow::WarningIcon,
                                    "Error !",
                                    "Could not write plot file.",
                                    "Ok");
      vcl_map<double,double>::const_iterator it;
      for (it = m_plot.begin(); it != m_plot.end(); ++it)
      {
        String line; line << it->first << " " << it->second << "\n";
        f.appendText(line);
      }
    }
  }
}
