/*
** distributions.h
** Login : <nicolas.burrus@ensta.fr>
** Started on  Thu Jun 23 23:26:21 2005 Nicolas Burrus
** $Id$
** 
** Copyright (C) 2005, 2006, 2007 Nicolas Burrus
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef   	NTK_STATS_DISTRIBUTIONS_H_
# define   	NTK_STATS_DISTRIBUTIONS_H_

# include <ntk/core.h>
# include <ntk/utils/debug.h>
# include <ntk/utils/math.h>

# include <vcl_memory.h>
# include <vcl_string.h>
# include <vcl_sstream.h>
# include <vcl_fstream.h>
# include <vcl_cfloat.h>
# include <vcl_vector.h>
# include <vcl_map.h>
# include <vcl_numeric.h>

# include <vnl/vnl_erf.h>
# include <vnl/vnl_random.h>

#define NTK_DISTRIB_CTOR_AND_PARAMS_1(Name, param1) \
  struct Params                                     \
  {                                                 \
    Params(double param1) : param1(param1) {}       \
    double param1;                                  \
  };                                                \
  Params p;                                         \
                                                    \
  Name(Params p) : p(p) {}                          \
  Name(double param1) : p(param1) {}
  
#define NTK_DISTRIB_CTOR_AND_PARAMS_2(Name, param1, param2)               \
  struct Params                                                           \
  {                                                                       \
    Params(double param1, double param2) : param1(param1), param2(param2) \
    {}                                                                    \
    double param1, param2;                                                \
  };                                                                      \
  Params p;                                                               \
                                                                          \
  Name(Params p) : p(p) {}                                                \
  Name(double param1, double param2) : p(param1, param2) {}

#define NTK_DISTRIB_CTOR_AND_PARAMS_3(Name, param1, param2, param3) \
  struct Params                                                     \
  {                                                                 \
    Params(double param1, double param2, double param3)             \
      : param1(param1), param2(param2), param3(param3)              \
    {}                                                              \
    double param1, param2, param3;                                  \
  };                                                                \
  Params p;                                                         \
                                                                    \
  Name(Params p) : p(p) {}                                          \
  Name(double param1, double param2, double param3)                 \
    : p(param1, param2, param3) {}


namespace ntk 
{

  unsigned draw_from_distrib(const vcl_vector<double>& distrib, 
                             vnl_random& rgen);
        
  template <class T, class U>
  double distrib_size(const vcl_map<T, U>& distrib)
  {
    typedef vcl_map<T, U> distrib_type;
    double nbelem = 0;
    foreach_const_it(it, distrib, typename distrib_type)
    { nbelem += it->second; }
    return nbelem;
  }
  
  inline
  void normalize_distrib(vcl_vector<double>& distrib)
  {
    double sum = vcl_accumulate(stl_bounds(distrib), 0.0);
    if (!flt_eq(sum, 0.0))
    {
      foreach_idx(i, distrib) distrib[i] /= sum;
    }
  }
  
  struct StudentDistrib
  {
    struct Params 
    { 
      Params(double degrees_of_freedom)
        : v(degrees_of_freedom)
      {}
      double v;
    };
    Params p;
    
    StudentDistrib(Params p) : p(p) {}
    StudentDistrib(double degrees_of_freedom)
      : p(degrees_of_freedom)
    {}
    
    struct InvCdf
    {
      InvCdf(Params p) : p(p) {}
      Params p;      
      double operator() (double x) const;
    };
    
    InvCdf invCdf() const { return InvCdf(p); }
  };
  
  struct WeibullDistrib
  {
    struct Params 
    { 
      Params(double beta, double etha, double gamma)
        : beta(beta), etha(etha), gamma(gamma)
      {}
      double beta, etha, gamma;
    };
    Params p;
    
    WeibullDistrib(Params p) : p(p) {}
    WeibullDistrib(double beta, double etha, double gamma) 
      : p(beta, etha, gamma) 
    {}
    
    struct Cdf
    {
      Cdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const
      { return 1.0-exp(-pow((x-p.gamma)/p.etha, p.beta)); }
    };
    
    struct InvCdf
    {
      InvCdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const
      { 
        ntk_assert(x <= 1.0, "Bad x");
        return p.gamma+p.etha*pow(log(1.0/(1-x)), (1/p.beta));
      }
    };
    
    struct Pdf
    {
      Pdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const
      { 
        return (p.beta/p.etha)*pow(x/p.etha, p.beta-1)
                *exp(-pow(x/p.etha, p.beta)); 
      }
    };
    
    struct LogPdf
    {
      LogPdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const      
      { 
        x = (x-p.gamma)/p.etha;
        return log(p.beta/p.etha) + (p.beta-1.0)*log(x) - pow(x, p.beta); 
      }
    };
    
    Cdf cdf() const { return Cdf(p); }
    InvCdf invCdf() const { return InvCdf(p); }
    Pdf pdf() const { return Pdf(p); }
    LogPdf logPdf() const { return LogPdf(p); }
  };
  
  inline double wblinv(double c, double beta, double etha, double gamma = 0)
  {
    WeibullDistrib m(beta, etha, gamma);
    return m.invCdf()(c);
  }
  
  struct ExpDistrib
  {
    struct Params
    {
      Params(double lambda) : lambda(lambda) {}
      double lambda;
    };
    Params p;
    
    ExpDistrib(double lambda)
      : p(lambda)
    {}
    
    struct Cdf
    {
      Cdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const
      { return 1.0-exp(-x/p.lambda); }
    };
            
    Cdf cdf() const { return Cdf(p); }
  };

  struct NormalDistrib
  {
    NTK_DISTRIB_CTOR_AND_PARAMS_2(NormalDistrib, mu, sigma);
    
    struct Cdf
    {
      Cdf(Params p) : p(p) {}
      Params p;
      double operator() (double x) const
      {
        double y = (x-p.mu)/p.sigma;
        return 0.5 * vnl_erfc(-y/sqrt(2.0));
      }
    };
      
    struct InvCdf
    {
      InvCdf(Params p) : p(p) {}
      Params p;
      double operator() (double x) const
      { return (-sqrt(2.0) * erfcinv(2.0*x))*p.sigma+p.mu; }
    };
    
    struct Pdf
    {
      Pdf(Params p) : p(p) {}
      Params p;      
      double operator() (double x) const
      { 
        double r = 1. / (p.sigma * sqrt(2.*M_PI));
        r *= exp(-(x-p.mu)*(x-p.mu) / (2. * p.sigma * p.sigma));
        return r;
      }
    };
    
    struct LogPdf
    {
      LogPdf(Params p) : p(p) {}
      Params p;      
      double operator() (double x) const      
      { 
        double a = (x-p.mu)/p.sigma;
        double pfa = 0.5*erfcx(-a/sqrt(2.0));
        pfa = log(pfa)-(a*a/2.);
        return pfa;
      }
    };
    
    struct Sampler
    {
      Sampler(Params p) : p(p) {}
      Params p;
      double operator() () const      
      { 
        double x1, x2, w, y1, y2;
        do 
        {
          x1 = 2.0f * rgen.drand64() - 1.0f;
          x2 = 2.0f * rgen.drand64() - 1.0f;
          w = x1 * x1 + x2 * x2;
        } while ( w >= 1.0f );
        w = sqrt( (-2.0 * log( w ) ) / w );
        y1 = x1 * w * p.sigma + p.mu;
        y2 = x2 * w * p.sigma + p.mu;
        return y1;
      }
      
    private:
        mutable vnl_random rgen;
    };
    
    Cdf cdf() const { return Cdf(p); }
    InvCdf invCdf() const { return InvCdf(p); }
    Pdf pdf() const { return Pdf(p); }
    LogPdf logPdf() const { return LogPdf(p); }
    Sampler sampler() const { return Sampler(p); }
  };
  
  struct LogNormalDistrib
  {
    NTK_DISTRIB_CTOR_AND_PARAMS_2(LogNormalDistrib, mu, sigma);
    
    struct Cdf
    {
      Cdf(Params p) : p(p) {}
      Params p;
      double operator() (double x) const
      { return 0.5 + 0.5*vnl_erf((log(x)-p.mu)/(p.sigma*sqrt(2.0))); }
    };
      
    struct InvCdf
    {
      InvCdf(Params p) : p(p) {}
      Params p;
      double operator() (double x) const
      { abort(); return 1.0; }
    };
    
    struct Pdf
    {
      Pdf(Params p) : p(p) {}
      Params p;      
      double operator() (double x) const
      { 
        return exp(-pow((log(x)-p.mu)/p.sigma, 2)/2.0)/(x*p.sigma*sqrt(2.0*M_PI)); 
      }
    };
    
    struct LogPdf
    {
      LogPdf(Params p) : p(p) {}
      Params p;      
      double operator() (double x) const      
      { 
        Pdf pdf(p);
        return log(pdf(x));
      }
    };
    
    Cdf cdf() const { return Cdf(p); }
    InvCdf invCdf() const { return InvCdf(p); }
    Pdf pdf() const { return Pdf(p); }
    LogPdf logPdf() const { return LogPdf(p); }
  };
  
  struct FisherTippettDistrib
  {
    NTK_DISTRIB_CTOR_AND_PARAMS_2(FisherTippettDistrib, mu, beta);
    
    struct Cdf
    {
      Cdf(Params p) : p(p) {}
      Params p;
      double operator() (double x) const
      { return exp(-exp(-(x-p.mu)/p.beta)); }
    };
      
    struct InvCdf
    {
      InvCdf(Params p) : p(p) {}
      Params p;
      double operator() (double x) const
      { abort(); return 1.0; }
    };
    
    struct Pdf
    {
      Pdf(Params p) : p(p) {}
      Params p;      
      double operator() (double x) const
      {
        double z = exp(-(x-p.mu)/p.beta);
        return (exp(-z)*z)/p.beta;
      }
    };
    
    struct LogPdf
    {
      LogPdf(Params p) : p(p) {}
      Params p;      
      double operator() (double x) const      
      {
        double z = exp(-(x-p.mu)/p.beta);
        return -z + log(z) - log(p.beta);
      }
    };
    
    Cdf cdf() const { return Cdf(p); }
    InvCdf invCdf() const { return InvCdf(p); }
    Pdf pdf() const { return Pdf(p); }
    LogPdf logPdf() const { return LogPdf(p); }
  };
  
  struct FrechetDistrib
  {
    NTK_DISTRIB_CTOR_AND_PARAMS_3(FrechetDistrib, mu, sigma, alpha);
    
    struct Cdf
    {
      Cdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const
      { 
        double z = (x-p.mu)/p.sigma;
        return exp(-pow(z, -p.alpha)); 
      }
    };
    
    struct InvCdf
    {
      InvCdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const
      { 
        abort(); return 0.;
      }
    };
    
    struct Pdf
    {
      Pdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const
      { 
        double z = (x-p.mu)/p.sigma;
        return (p.alpha/p.sigma)*pow(z, -p.alpha-1.0)*exp(-pow(z, -p.alpha)); 
      }
    };
    
    struct LogPdf
    {
      LogPdf(Params p) : p(p) {}
      Params p;
      
      double operator() (double x) const      
      {
        ntk_assert((x-p.mu) >= 0, "Bad x.");
        // if ((x-p.mu) < 0) return -FLT_MAX;
        double z = (x-p.mu)/p.sigma;
        double r = log(p.alpha/p.sigma) + (-p.alpha-1.0)*log(z) - pow(z, -p.alpha); 
        return r;
      }
    };
    
    Cdf cdf() const { return Cdf(p); }
    InvCdf invCdf() const { return InvCdf(p); }
    Pdf pdf() const { return Pdf(p); }
    LogPdf logPdf() const { return LogPdf(p); }
  };
  
} // end of ntk

#endif	    /* !NTK_STATS_DISTRIBUTIONS_H_ */
