/*
** math.h
** Login : <nicolas.burrus@ensta.fr>
** Started on  Thu Jun 23 23:26:21 2005 Nicolas Burrus
** $Id$
**
** Copyright (C) 2005, 2006, 2007 Nicolas Burrus
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef   	NTK_UTILS_MATH_H_
# define   	NTK_UTILS_MATH_H_

# include <ntk/core.h>
# include <vnl/vnl_erf.h>

#define NTK_UNSAFE_MAX(a,b) ((a)<(b)?(b):(a))

namespace ntk {

  // Float comparisons.

  inline bool
  flt_eq (float lhs, float rhs,
          float epsilon = vcl_numeric_limits<float>::epsilon())
  {
    return vcl_abs(lhs - rhs) <= epsilon;
  }

  inline bool
  flt_leq (float lhs, float rhs,
           float epsilon = vcl_numeric_limits<float>::epsilon())
  {
    return lhs <= (rhs+epsilon);
  }

  inline bool
  flt_geq (float lhs, float rhs,
           float epsilon = vcl_numeric_limits<float>::epsilon())
  {
    return (lhs+epsilon) >= rhs;
  }

  template <class T, class U>
  inline bool
  in_range(T lower, U data, T upper)
  {
    return (data >= lower) && (data <= upper);
  }

  inline unsigned
  add_and_saturate(unsigned a, unsigned b, unsigned max_value)
  {
    if ((a + b) > max_value)
      return max_value;
    return a + b;
  }

  double normalize_angle_0_2pi(double angle);
  inline double rad_to_deg(double rad_angle) { return (rad_angle*180.0)/M_PI; }
  inline double deg_to_rad(double deg_angle) { return (deg_angle*M_PI)/180.0; }

  double erfcx(double x);

  // From codecogs : http://www.codecogs.com/
  double erfinv(double y);
  inline double erfcinv(double x) { return erfinv(1.0-x); }


  inline double norminv(double p) { return -sqrt(2.0) * erfcinv(2.0*p); }
  inline double norminv(double p, double mean, double dev) { return norminv(p)*dev + mean; }
  double normcdf(double x, double mean, double std_dev);
  double normpdf(double x, double mean, double dev);
  double lnnormcdf(double x, double mean, double std_dev);

  double lnchoosek(unsigned n, unsigned k);

  inline double sigmoid(double x) { return 1.0/(1.0+exp(-x)); }

  // Returns an upper bound of log(B(k, n, p))
  double binomial_logicdf_hoeffding_estimate(uint32 k, double p, uint32 n);

#ifdef USE_GSL
  double binomial_logcdf_gsl(uint32 k, double p, uint32 n);
  double binomial_logicdf_gsl(uint32 k, double p, uint32 n);
  double binomial_logicdf_best_estimate(uint32 k, double p, uint32 n);
#endif // def USE_GSL

  
  inline double recursive_mean(double previous_mean, double previous_size, double new_value)
  {
    if (previous_size < 0.01) return new_value;

    double t = previous_size + 1;
    return (previous_mean * (t-1) + new_value)/t;
  }
  
} // end of ntk

#endif	    /* !NTK_UTILS_MATH_H_ */
