//
// stl.h
// 
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef NTK_UTILS_stl_H
# define NTK_UTILS_stl_H

# include <ntk/core.h>
# include <ntk/utils/debug.h>

# include <vcl_set.h>
# include <vcl_algorithm.h>
# include <vcl_iterator.h>

template <class T>
vcl_set<T> operator+(const vcl_set<T>& lhs, const vcl_set<T>& rhs)
{
  vcl_set<T> result;
  set_union(stl_bounds(lhs), stl_bounds(rhs), 
            vcl_inserter(result, result.begin()));
  return result;
}

template <class T>
vcl_set<T>& operator+=(vcl_set<T>& lhs, const vcl_set<T>& rhs)
{
  lhs.insert(stl_bounds(rhs));
  return lhs;
}

namespace ntk
{
  
  template <class ContainerType>
  bool has_key(const ContainerType& container, 
               const typename ContainerType::key_type& key)
  {
    return container.find(key) != container.end();
  }
   
  template <class Container>
  class AssociativeContainerTransaction
  {
    public:
      typedef typename Container::mapped_type mapped_type;
      typedef typename Container::key_type key_type;
    
    public:
      AssociativeContainerTransaction(Container& container)
        : m_container(container)
      {
      }
    
      ~AssociativeContainerTransaction()
      { validateTransaction(); }
      
      const mapped_type& operator[] (const key_type& key)
      { 
        typename Container::const_iterator it = m_container.find(key);
        ntk_ensure(it != m_container.end(), "Inexistant key.");
        return it->second;
      }
      
      mapped_type& changeElement(const key_type& key)
      {
        typename Container::iterator it = m_container.find(key);
        ntk_ensure(it != m_container.end(), "Inexistant key.");
        if (!has_key(m_old_values, it->first))
          m_old_values[it->first] = it->second;
        return it->second;
      }
      
      void removeElement(const key_type& key)
      {
        typename Container::iterator it = m_container.find(key);
        ntk_ensure(it != m_container.end(), "Inexistant key.");
        if (!has_key(m_old_values, it->first))
          m_old_values[it->first] = it->second;
        m_container.erase(key);
      }
      
      void abortTransaction()
      {
        for(typename Container::const_iterator it = m_old_values.begin();
            it != m_old_values.end();
            ++it)
          m_container[it->first] = it->second;
      }
      
      void validateTransaction()
      {
        m_old_values.clear();
      }
      
    private:
      Container& m_container;
      Container m_old_values;
  };
  
  template <class Container>
  AssociativeContainerTransaction<Container>
  create_associative_container_transaction(Container& container)
  {
    return AssociativeContainerTransaction<Container>(container);
  }
  
  template <class Iterator>
  class key_iterator : public Iterator
  {
    typedef typename vcl_iterator_traits<Iterator>::value_type value_type;
    
    public:
      key_iterator(const Iterator& it)
        : Iterator(it)
      {}

      const typename value_type::first_type& operator*() const 
      { return (*(Iterator*)this)->first; }
      
      const typename value_type::first_type* operator->() const 
      { return &(*(Iterator*)this)->first; }
  };
    
  template <class Iterator>
  class value_iterator : public Iterator
  {
    typedef typename vcl_iterator_traits<Iterator>::value_type value_type;
    
    public:
      value_iterator(const Iterator& it)
      : Iterator(it)
      {}

      const typename value_type::second_type& operator*() const 
      { return (*(Iterator*)this)->second; }
      
      const typename value_type::second_type* operator->() const 
      { return &(*(Iterator*)this)->second; }
  };
  
} // end of ntk

#endif // ndef NTK_UTILS_stl_H
