/*
** vil_binary_image.cc
** Login : <nicolas.burrus@ensta.fr>
** Started on  Thu Jun 23 23:35:30 2005 Nicolas Burrus
** $Id$
** 
** Copyright (C) 2005, 2006 Nicolas Burrus
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "vil_binary_image.h"
#include <ntk/vxl/vil_utils.h>

VIL_SMART_PTR_INSTANTIATE(ntk::vil_dynamic_bitset_ptr);

namespace ntk {

  vil_binary_image& 
  vil_binary_image::operator= (const vil_binary_image& rhs)
  {
    ni_ = rhs.ni_;
    nj_ = rhs.nj_;
    nplanes_ = rhs.nplanes_;
    istep_ = rhs.istep_;
    jstep_ = rhs.jstep_;
    bitsets_sptr_ = rhs.bitsets_sptr_;
    bitsets_ = rhs.bitsets_;
    return *this;
  }

  void
  vil_binary_image::load (const char* filename)
  {
    vil_image_view<bool> vim = vil_load(filename);
    deep_copy(vim);
    flip();
  }

  void
  vil_binary_image::deep_copy(const vil_binary_image& src)
  {
    set_size(src.ni(), src.nj(), src.nplanes());
    for (unsigned p = 0; p < nplanes_; ++p)
      bitsets_[p] = src.bitsets_[p];
  }

  //: resize to width x height x n_planes.
  // If already correct size, this function returns quickly
  void
  vil_binary_image::set_size(unsigned width, 
                             unsigned height, 
                             unsigned n_planes)
  {
    if (ni_ == width && nj_ == height && nplanes_ == n_planes)
      return;

    ni_ = width;
    nj_ = height;
    nplanes_ = n_planes;
    istep_ = 1;
    jstep_ = width;
    release_memory();
    bitsets_sptr_ = new vil_dynamic_bitset_ptr(n_planes, width*height);
    bitsets_ = bitsets_sptr_->data();
  };

  unsigned
  vil_binary_image::count () const
  { 
    unsigned sum = 0;
    for (uint32 p = 0; p < nplanes_; ++p)
      sum += bitsets_[p].count();
    return sum;
  }

  void
  vil_binary_image::fill (bool value)
  {    
    if (value)
      for (uint32 p = 0; p < nplanes_; ++p)
        bitsets_[p].set();
    else
      for (uint32 p = 0; p < nplanes_; ++p)
        bitsets_[p].reset();
  }

  vil_binary_image& 
  vil_binary_image::operator|= (const vil_binary_image& rhs)
  {
    for (unsigned p = 0; p < nplanes_; ++p)
      bitsets_[p] |= rhs.bitsets_[p];
    return *this;
  }
  
  vil_binary_image& 
  vil_binary_image::operator&= (const vil_binary_image& rhs)
  {
    for (unsigned p = 0; p < nplanes_; ++p)
      bitsets_[p] &= rhs.bitsets_[p];
    return *this;
  }

  vil_binary_image& 
  vil_binary_image::operator^= (const vil_binary_image& rhs)
  {
    for (unsigned p = 0; p < nplanes_; ++p)
      bitsets_[p] ^= rhs.bitsets_[p];
    return *this;
  }

  bool
  vil_binary_image::operator== (const vil_binary_image& rhs) const
  { 
    for (unsigned p = 0; p < nplanes_; ++p)
      if (bitsets_[p] != rhs.bitsets_[p])
        return false;
    return true;
  }

  vil_binary_image 
  operator| (const vil_binary_image& lhs, const vil_binary_image& rhs)
  {
    vil_binary_image out; out.deep_copy(lhs);
    out |= rhs;
    return out;
  }

  vil_binary_image 
  operator& (const vil_binary_image& lhs, const vil_binary_image& rhs)
  {
    vil_binary_image out; out.deep_copy(lhs);
    out &= rhs;
    return out;
  }

  vil_binary_image 
  operator^ (const vil_binary_image& lhs, const vil_binary_image& rhs)
  {
    vil_binary_image out; out.deep_copy(lhs);
    out ^= rhs;
    return out;
  }

  bool 
  vil_save(const vil_binary_image &im, char const* filename)
  {
    // FIXME: this is a workaround for a vil bug.
    vil_binary_image tmp; tmp.deep_copy(im); tmp.flip();
    vil_image_view<bool> view_im;
    tmp.copy_to(view_im);
    return vil_save(view_im, filename);
  }

  vil_binary_image
  vil_plane(const vil_binary_image& im, unsigned nplane)
  {
    vil_binary_image out (im.bitsets_sptr(), im.bitsets()+nplane, 
                          im.ni(), im.nj(), 1);
    return out;
  }
  
  template <bool critical_value_>
  class vil_compute_morpho_value
  {
  public:
    vil_compute_morpho_value (const vil_binary_image& src_image,
                              const vil_structuring_element& element,
                              vil_border<vil_binary_image> border)
      : element_(element),
        src_image_(src_image),
        border_(border)
    {
      vil_compute_offsets(offset_,element,src_image.istep(),src_image.jstep());
    }

    inline
    bool apply_at (unsigned i, unsigned j) const
    {    
      for (uint32 k = 0; k < element_.p_i().size(); ++k)
      {
        if (border_(src_image_,
                    i+element_.p_i()[k], 
                    j+element_.p_j()[k]) == critical_value_)
          return critical_value_;
      }
      return !critical_value_;
    }

    inline
    bool apply_at (int index) const
    {
      assert(border_.is_constant());
      const bool constant_value = border_.constant_value();
      const int max_index = src_image_.size();
      int new_index;
      bool val;
      for (uint32 i = 0; i < offset_.size(); ++i)
      {
        new_index = index + offset_[i];
        if (new_index < 0 || new_index >= max_index)
          val = constant_value;
        else
          val = src_image_[new_index];
        if (val == critical_value_)
          return critical_value_;
      }
      return !critical_value_;
    }

    inline
    bool apply_noborder_at (int index) const
    {
      const int max_index = src_image_.size();
      int new_index;
      for (uint32 i = 0; i < offset_.size(); ++i)
      {
        new_index = index + offset_[i];
        assert (new_index >= 0 && new_index < max_index);
        if (src_image_[new_index] == critical_value_)
          return critical_value_;
      }
      return !critical_value_;
    }

  private:
    const vil_structuring_element& element_;
    const vil_binary_image& src_image_;
    vil_border<vil_binary_image> border_;
    vcl_vector<vcl_ptrdiff_t> offset_;
  };

  template <class computerT>
  void vil_binary_morpho(const vil_binary_image& src_image, 
                         vil_binary_image& dest_image,
                         const vil_structuring_element& element,
                         const vil_border<vil_binary_image>& border,
                         const computerT& computer)
  {
    assert(src_image.nplanes()==1);
    unsigned ni = src_image.ni();
    unsigned nj = src_image.nj();
    dest_image.set_size(ni,nj,1);
    dest_image.fill(0);

    vcl_ptrdiff_t s_istep = src_image.istep(),  s_jstep = src_image.jstep();
    vcl_ptrdiff_t d_istep = dest_image.istep(), d_jstep = dest_image.jstep();

    // Define box in which all elements will be valid
    int ilo = -element.min_i();
    int ihi = ni-1-element.max_i();
    int jlo = -element.min_j();
    int jhi = nj-1-element.max_j();

    if (border.is_constant())
    {
      const vil_binary_image::bitset_type& src_bitset = src_image.bitset();
      vil_binary_image::bitset_type& dest_bitset = dest_image.bitset();
      
      const unsigned n_pos = src_bitset.size();
      
      unsigned cur_pos = src_bitset.find_next(0);
      while (cur_pos < n_pos)
      {
        dest_bitset[cur_pos] = computer.apply_at(cur_pos);
        cur_pos = src_bitset.find_next(cur_pos);
      }      
    }
    else // Handle border
    {
      // Deal with left edge
      for (unsigned j=0; j<nj; ++j)
      for (int i=0; i<ilo; ++i)
        dest_image.set(i,j, computer.apply_at(i,j));
    
      // Deal with right edge
      for (unsigned j=0; j<nj; ++j)
      for (unsigned i=ihi+1;i<ni;++i)
        dest_image.set(i,j, computer.apply_at(i,j));
    
      // Deal with bottom edge
      for (int j=0; j<jlo; ++j)
      for (unsigned i=0; i<ni; ++i)
        dest_image.set(i,j, computer.apply_at(i,j));
    
      // Deal with top edge
      for (unsigned j=jhi+1; j<nj; ++j)
      for (unsigned i=0; i<ni; ++i)
        dest_image.set(i,j, computer.apply_at(i,j));

      // Slow implementation.
      for (int j=jlo;j<=jhi;++j)
      {
        unsigned src_p = j*s_jstep + ilo*s_istep;
        unsigned dest_p = j*d_jstep + ilo*d_istep;
        
        for (int i=ilo;i<=ihi;++i,src_p+=s_istep,dest_p+=d_istep)
          dest_image.set(dest_p, computer.apply_noborder_at(src_p));
      }
    }
  }

  void
  vil_binary_erode(const vil_binary_image& src_image, 
                   vil_binary_image& dest_image,
                   const vil_structuring_element& element)
  {
    vil_border<vil_binary_image> border = 
      vil_border_create_constant(src_image, true);
    vil_binary_erode(src_image, dest_image, element, border);
  }

  void
  vil_binary_erode(const vil_binary_image& src_image, 
                   vil_binary_image& dest_image,
                   const vil_structuring_element& element,
                   vil_border<vil_binary_image> border)
  {
    vil_compute_morpho_value<0> eroder(src_image, element, border);
    vil_binary_morpho(src_image, dest_image, element, border, eroder);
  }

  void
  vil_binary_dilate(const vil_binary_image& src_image, 
                    vil_binary_image& dest_image,
                    const vil_structuring_element& element)
  {
    vil_border<vil_binary_image> border = 
      vil_border_create_constant(src_image, false);
    vil_binary_dilate(src_image, dest_image, element, border);
  }

  void
  vil_binary_dilate(const vil_binary_image& src_image, 
                    vil_binary_image& dest_image,
                    const vil_structuring_element& element,
                    vil_border<vil_binary_image> border)
  {
    assert(border.is_constant());
    assert(src_image.nplanes()==1);
    assert(border.constant_value() == 0);

    unsigned ni = src_image.ni();
    unsigned nj = src_image.nj();
    dest_image.set_size(ni,nj,1);
    dest_image.deep_copy(src_image);

    vcl_ptrdiff_t s_istep = src_image.istep(),  s_jstep = src_image.jstep();
    vcl_ptrdiff_t d_istep = dest_image.istep(), d_jstep = dest_image.jstep();

    vcl_vector<vcl_ptrdiff_t> offset;
    vil_compute_offsets(offset,element,src_image.istep(),src_image.jstep());
    
    const vil_binary_image::bitset_type& src_bitset = src_image.bitset();
    vil_binary_image::bitset_type& dest_bitset = dest_image.bitset();
    
    const unsigned n_pos = src_bitset.size();
    
    int offset_size = offset.size();
    int max_index = src_image.size();
    int new_pos;
    unsigned cur_pos = src_bitset.find_next(0);
    while (cur_pos < n_pos)
    {
      for (int i = 0; i < offset_size; ++i)
      {
        // It is minus because we have to invert the structuring element.
        new_pos = cur_pos - offset[i];
        if (new_pos < 0 || new_pos >= max_index)
          continue;
        dest_bitset[new_pos] = true;
      }
      cur_pos = src_bitset.find_next(cur_pos);
    }
  }

} // end of ntk
