//
// ac_learn_process.h
// 
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef ACSEGMENTOR_ac_learn_process_H
# define ACSEGMENTOR_ac_learn_process_H

# include <ntk/ntk.h>
# include <vcl_set.h>

class AcSegmentor; ntk_ptr_typedefs(AcSegmentor);

class AcLearner
{
  public:
    AcLearner(uint32 nb_categories)
      : m_thresholds(nb_categories,-FLT_MAX), m_epsilon(1)
    {}
  
    virtual ~AcLearner() {}
    
    void initialize() { initializeImpl(); }
    virtual void initializeImpl() = 0;
    
  public:
    void setEpsilon(double eps) { m_epsilon = eps; }
    const vcl_vector<double>& getThresholds() const { return m_thresholds; }
    uint32 nbCategories() const { return m_thresholds.size(); }
  
  public:
    virtual void newImage() = 0;
    virtual void addDistance(uint32 couple_category, double value) = 0;
    virtual void computeThresholds() = 0;
      
  protected:
    vcl_vector<double> m_thresholds;
    double m_epsilon;
};
ntk_ptr_typedefs(AcLearner);

class AcMaxLearner : public AcLearner
{
  public:
    AcMaxLearner(uint32 nb_categories)
      : AcLearner(nb_categories),
        m_raw_values(nb_categories)
     {}
  
    virtual void initializeImpl();
        
  public:
    virtual void newImage() {}
    virtual void addDistance(uint32 couple_category, double value);
    virtual void computeThresholds();
    
  private:
    void completeMissingValues();
    
  private:
    vcl_vector<vcl_vector<double> > m_raw_values;
};

class AcMeanLearner : public AcLearner
{
  public:
    struct MaxImageValues
    {
      MaxImageValues(int nb_index = 0) : values(nb_index) {}
      uint32 number_of_values_to_keep() const { return 10; }
      void addSorted(uint32 idx, double value);
      vcl_vector< vcl_multiset<double> > values;
    };
  
  public:
    AcMeanLearner(uint32 nb_categories)
      : AcLearner(nb_categories),
        m_raw_values(nb_categories)
    {}
  
    virtual void initializeImpl();
        
  public:
    virtual void newImage();
    virtual void addDistance(uint32 couple_category, double value);
    virtual void computeThresholds();        
    
  private:
    void optimizeThreshold(uint32 k);
    
  private:
    vcl_vector<MaxImageValues> m_raw_values;
};

class AcLearnProcess
{
  public:
    AcLearnProcess(const PropertySet& properties, AcSegmentorPtr segmentor)
      : m_segmentor(segmentor), m_listener(0), m_should_stop(false)
    {
      m_width = properties.getIntValue(JUCE_T("learning-width"));
      m_height = properties.getIntValue(JUCE_T("learning-height"));
      m_nb_iterations = properties.getIntValue(JUCE_T("learning-iterations"));
      m_nb_categories = properties.getIntValue(JUCE_T("learning-categories"));
      m_learner = createLearner(properties);
      m_current_thresholds.resize(m_nb_categories, -FLT_MAX);
    }

  public:
    void setShouldStop() 
    { 
      const ScopedLock lock(m_critical_section); 
      m_should_stop = true; 
    }
    bool shouldStop() const
    { 
      const ScopedLock lock(m_critical_section); 
      return m_should_stop;
    }
    
    uint32 nbCategories() const { return m_nb_categories; }

    void setChangeListener(ChangeListener* listener) { m_listener = listener; }

    const AcSegmentorPtr& getSegmentor() const { return m_segmentor; }

    ntk::vil_byte_image getCurrentSourceImage() const
    { 
      const ScopedLock lock(m_critical_section); 
      return m_source_image;
    }
      
    vcl_vector<double> getCurrentThresholds() const 
    { 
      const ScopedLock lock(m_critical_section); 
      return m_current_thresholds;
    }    
    
  public:
    void learn();

  protected:
    void saveResultsToDatabase();
    AcLearnerPtr createLearner(const PropertySet& properties);
    
  private:
    ntk::vil_byte_image m_source_image;
    uint32 m_width;
    uint32 m_height;
    uint32 m_nb_iterations;
    uint32 m_nb_categories;
    AcSegmentorPtr m_segmentor;
    AcLearnerPtr m_learner;
    ChangeListener* m_listener;
    CriticalSection m_critical_section;
    bool m_should_stop;
    vcl_vector<double> m_current_thresholds;
};
ntk_ptr_typedefs(AcLearnProcess);

#endif // ndef ACSEGMENTOR_ac_learn_process_H
